/*
    REPLICATION OF THE DEMOGRAPHIC PRISONER'S DILEMMA
    Copyright (C) <2009>  <Wolfgang Radax and Bernhard Rengs>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Corresponding author: Wolfgang Radax <wolfgang.radax@econ.tuwien.ac.at>
 */

package demographicpd;

import java.util.List;

import uchicago.src.sim.gui.Drawable;
import uchicago.src.sim.gui.SimGraphics;

import java.awt.*;
import java.awt.Point;

/**
 * Agent implements all the activities, agents in the demographic prisoner's dilemma
 * are supposed to conduct: move, play, give birth.
 * @author Wolfgang Radax, Vienna University of Technology
 */
public class Agent implements Drawable {
    
    private int             id;
    private int             x, y;
    private int             resources;
    private int             age;
    private int             behavior;
    
    private boolean         isAlive;
    private boolean         newborn;
    
    private DemographicPD   dpd;
    
    /**
     * Default constructor. It is not possible to create an agent object without
     * specifying her coordinates, her id-number and a reference to the model itself.
     * @param id    id-number of the agent
     * @param x     x-coordinate of the agent's current location
     * @param y     y-coordinate of the agent's current location
     * @param dpd   reference to the model
     */
    public Agent(int id, int x, int y, DemographicPD dpd) {
        this.id = id;
        this.x = x;
        this.y = y;
        this.resources = dpd.getInitResources();

        this.age = 0;
        if (dpd.getRandomBirthAge()) {
            if (dpd.getMaxAge() > 0) {
                age = dpd.nextRandomIntFromTo(0, dpd.getMaxAge() - 1);
            }
        }

        this.behavior = PrisonersDilemma.COOPERATE;
        this.isAlive = true;
        this.newborn = false;
        this.dpd = dpd;
    }
    
    private boolean isProlific() {
        boolean prolific = false;
        
        if (isAlive() //agent is alive
                && !isNewBorn() //not born in the current period
                && age >= dpd.getBirthMinAge() //at least as old as the min required proliferation age
                && resources >= dpd.getBirthResources()) { //and wealthy enough

            if (dpd.getBirthMaxAge() > 0) {
                //maximum proliferation age specified
                if (age <= dpd.getBirthMaxAge()) {
                    prolific = true;
                }
            }
            else {
                //no maximum proliferation age specified
                prolific = true;
            }
        }
        return prolific;
    }
    
    /**
     * Lets the agent move to a randomly chosen unoccupied field within his
     * Von-Neumann neighborhood. If every field is occupied, the agent doesn't move.
     */
    public void move() {
        if (isAlive()) {
            Point p = getRandomEmptyNeighborhoodSpot();
            if (p != null) {
                dpd.getSpace().putObjectAt(x, y, null);
                dpd.getSpace().putObjectAt((int) p.getX(), (int) p.getY(), this);
                x = (int) p.getX();
                y = (int) p.getY();
            }
        }
    }
    
    /**
     * Lets the agent play against agent <code>n</code> ("neighbor"). Both agents
     * "choose" a strategy according to their type (cooperator or defector), then
     * the payoffs are evaluated and accumulated.
     * @param n Opponent against which the agent plays a game of PD
     */
    public void play(Agent n) {
        if (isAlive()) {
            int agentPayoff;
            int neighborPayoff;
            
            PrisonersDilemma pd = dpd.getPrisonersDilemma();

            agentPayoff = pd.getPayoff(getBehavior(), n.getBehavior());
            neighborPayoff = pd.getPayoff(n.getBehavior(), getBehavior());

            setResources(getResources() + agentPayoff);
            if (getResources() < 0) {
                //agent dies
                setAlive(false);
                if (dpd.getRemoveDeadImmediately()) {
                    dpd.space.putObjectAt(x, y, null);
                }
            }

            n.setResources(n.getResources() + neighborPayoff);

            if (dpd.getDieImmediately()) {
                if (n.getResources() < 0) {
                    //neighbor dies
                    n.setAlive(false);
                    if (dpd.getRemoveDeadImmediately()) {
                        dpd.space.putObjectAt(n.getX(), n.getY(), null);
                    }
                }
            }
        }
    }
    
    /**
     * Lets the agent give birth to offspring on a (Von Neumann-)neighboring
     * unoccupied site if the agent fulfills all necessary conditions (minimum
     * birth age, maximum birth age, enough resources). If all sites in the
     * agent's neighborhood are occupied, he can't give birth.
     * @param   newId id-number of new born agent
     * @return  New born agent in the case of successful birth, <code>null</code>
     *          otherwise.
     */
    public Agent giveBirth(int newId) {
        if (isProlific()) {
            Point p = getRandomEmptyNeighborhoodSpot();
            if (p != null) {
                Agent newAgent;
                newAgent = new Agent(newId, (int) p.getX(), (int) p.getY(), dpd);
                newAgent = mutate(newAgent);
                newAgent.setNewBorn(true);

                if (dpd.getInitialEndowmentInherited()) {
                    resources -= newAgent.getResources();
                }
                return newAgent;
            }
        }
        return null;
    }
    
    private Agent mutate(Agent newAgent) {
        int mutation = 0;
        if (dpd.getMutationProbability() > 0) {
            mutation = dpd.nextRandomIntFromTo(1, 100);
            if (mutation <= dpd.getMutationProbability()) {
                if (behavior == PrisonersDilemma.COOPERATE) {
                    newAgent.setBehavior(PrisonersDilemma.DEFECT);
                }
                else {
                    newAgent.setBehavior(PrisonersDilemma.COOPERATE);
                }
            }
            else {
                newAgent.setBehavior(behavior);
            }
        }
        else {
            newAgent.setBehavior(getBehavior());
        }
        return newAgent;
    }
    
    private Point getRandomEmptyNeighborhoodSpot() {
        int emptySpaces = 0;
        int[] spaces = new int[4];
        Point p;
        
        if (dpd.getSpace().getObjectAt(x, (y - 1 + dpd.getSpaceYSize()) % dpd.getSpaceYSize()) == null) {
            //up is empty
            spaces[emptySpaces] = 1;
            emptySpaces++;                    
        }
        if (dpd.getSpace().getObjectAt(x, (y + 1) % dpd.getSpaceYSize()) == null) {
            //down is empty
            spaces[emptySpaces] = 2;
            emptySpaces++;
        }
        if (dpd.getSpace().getObjectAt((x - 1 + dpd.getSpaceXSize()) % dpd.getSpaceXSize(), y) == null) {
            //left is empty
            spaces[emptySpaces] = 3;
            emptySpaces++;
        }
        if (dpd.getSpace().getObjectAt((x + 1) % dpd.getSpaceXSize(), y) == null) {
            //right is empty
            spaces[emptySpaces] = 4;
            emptySpaces++;
        }
        
        int newX = x, newY = y;
        
        if (emptySpaces > 0) {
            int moveTo = spaces[dpd.nextRandomIntFromTo(0, emptySpaces - 1)];
            
            if (moveTo == 1) {
                //move up
                newY = (y - 1 + dpd.getSpaceYSize()) % dpd.getSpaceYSize();
            }
            else if (moveTo == 2) {
                newY = (y + 1) % dpd.getSpaceYSize();
            }
            else if (moveTo == 3) {
                //move left
                newX = (x - 1 + dpd.getSpaceXSize()) % dpd.getSpaceXSize();
            }
            else if (moveTo == 4) {
                //move right
                newX = (x + 1) % dpd.getSpaceXSize();
            }
            
            p = new Point(newX, newY);
        }
        else {
            // neighborhood occupied, return null
            p = null;
        }

        return p;
    }
    
    public final void draw(final SimGraphics g) {
        final int wScale = g.getCellWidthScale();
        final int hScale = g.getCellHeightScale();
        
        g.setDrawingParameters(wScale, hScale, 0);
        
        Color c;
        if (this.isDefector()) {
            c = new Color(Math.max(0,Math.min(255, resources + 50)), 0, 0);
        }
        else {
            c = new Color(0, Math.max(0,Math.min(255, resources + 50)), 0);
        }
        
        g.drawFastRect(c);
        
    }
    
    //SETTERS AND GETTERS
    final List<Agent> getNeighbors() {
        return (dpd.getSpace()).getVonNeumannNeighbors(x, y, false);
    }
    
    public int getId() {
        return id;
    }
    
    public int getX() {
        return x;
    }
    
    public int getY() {
        return y;
    }
    
    /**
     * Determines the current x-coordinate of the agent.
     * @param x x-coordinate of the agent
     */
    public void setX(int x) {
        this.x = x;
    }
    
    /**
     * Determines the current y-coordinate of the agent.
     * @param y y-coordinate of the agent
     */
    public void setY(int y) {
        this.y = y;
    }
    
    public int getResources() {
        return resources;
    }
    
    /**
     * Determines the amount of resources of the agent.
     * @param res amount of resources of the agent
     */
    public void setResources (int res) {
        this.resources = res;
    }
    
    /**
     * Determines the behavior ({@link PrisonersDilemma#COOPERATE} or {@link PrisonersDilemma#DEFECT}) of the agent.
     * @param behavior behavior of the agent
     */
    public void setBehavior(int behavior) {
        this.behavior = behavior;
    }
    
    public int getBehavior() {
        return this.behavior;
    }
    
    public boolean isDefector() {
        return (this.behavior == PrisonersDilemma.DEFECT);
    }
    
    /**
     * Determines the current age of the agent
     * @param age current age of the agent
     */
    public void setAge(int age) {
        this.age = age;
    }
    
    public int getAge() {
        return age;
    }
    
    public boolean isAlive() {
        return this.isAlive;
    }
    
    /**
     * Flags the agent as alive or dead.
     * @param alive
     */
    public void setAlive(boolean alive) {
        this.isAlive = alive;
    }
    
    public boolean isNewBorn() {
        return this.newborn;
    }
    
    /**
     * Flags the agent as new born (born in the current period) or not.
     * @param newBorn
     */
    public void setNewBorn(boolean newBorn) {
        this.newborn = newBorn;
    }
}
