/*
    REPLICATION OF THE DEMOGRAPHIC PRISONER'S DILEMMA
    Copyright (C) <2009>  <Wolfgang Radax and Bernhard Rengs>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Corresponding author: Wolfgang Radax <wolfgang.radax@econ.tuwien.ac.at>
 */

package demographicpd;

import java.util.ArrayList;

/**
 * AgentList is an extended version of ArrayList to store all agent objects of the simulation.
 * Inherits from ArrayList and additionally provides several reporters for aggregate statistics.
 * @author Wolfgang Radax, Vienna University of Technology
 */
public class AgentList extends ArrayList<Agent> {

    /**
     * Default constructor. Calls constructor of super-class.
     */
    public AgentList() {
        super();
    }
    
    //REPORTERS
    /**
     * Returns the number of defectors.
     * @return Number of defectors
     */
    public int getNumberDefectors() {
        int numDefectors = 0;
        for (Agent a : this) {
            if (a.isDefector()) {
                numDefectors++;
            }
        }
        return numDefectors;
    }
    
    /**
     * Returns the number of cooperators.
     * @return Number of cooperators
     */
    public int getNumberCooperators() {
        int numCooperators = 0;
        for (Agent a : this) {
            if (!a.isDefector()) {
                numCooperators++;
            }
        }
        return numCooperators;
    }

    /**
     * Returns the average population age.
     * @return Average age of the population
     */
    public double getAvgAge() {
        int currNumAgents = 0;
        int sumAges = 0;
        for (Agent a : this) {
            sumAges += a.getAge();
            currNumAgents++;
        }
        return (double) sumAges / currNumAgents;
    }
    
    /**
     * Returns the average age of all defectors.
     * @return Average age of all defectors
     */
    public double getAvgDefectorAge() {
        int currNumAgents = 0;
        int sumAges = 0;
        for (Agent a : this) {
            if (a.isDefector()) {
                sumAges += a.getAge();
                currNumAgents++;
            }
        }
        return (double) sumAges / currNumAgents;
    }
    
    /**
     * Returns the average age of all cooperators.
     * @return Average age of all cooperators
     */
    public double getAvgCooperatorAge() {
        int currNumAgents = 0;
        int sumAges = 0;
        for (Agent a : this) {
            if (!a.isDefector()) {
                sumAges += a.getAge();
                currNumAgents++;
            }
        }
        return (double) sumAges / currNumAgents;
    }
    
    /**
     * Returns the average resources of the population.
     * @return Average resources of population
     */
    public double getAvgResources() {
        int currNumAgents = 0;
        int sumResources = 0;
        for (Agent a : this) {
            sumResources += a.getResources();
            currNumAgents++;
        }
        return (double) sumResources / currNumAgents;
    }
    
    /**
     * Returns the average resources of all defectors.
     * @return Average resources of all defectors
     */
    public double getAvgDefectorResources() {
        int currNumAgents = 0;
        int sumResources = 0;
        for (Agent a : this) {
            if (a.isDefector()) {
                sumResources += a.getResources();
                currNumAgents++;
            }
        }
        return (double) sumResources / currNumAgents;
    }
    
    /**
     * Returns the average resources of all cooperators.
     * @return Average resources of all cooperators
     */
    public double getAvgCooperatorResources() {
        int currNumAgents = 0;
        int sumResources = 0;
        for (Agent a : this) {
            if (!a.isDefector()) {
                sumResources += a.getResources();
                currNumAgents++;
            }
        }
        return (double) sumResources / currNumAgents;
    }
    
    /**
     * Returns the current share of defectors in the population.
     * @return Share of defectors in the population
     */
    public double getCurrentDefectorRatio() {
        int currNumAgents = 0;
        int currNumDefectors = 0;
        for (Agent a : this) {
            if (a.isDefector()) {
                currNumDefectors++;
            }
            currNumAgents++;
        }
        return (double) currNumDefectors*100 / currNumAgents;
    }
}
