/*
    REPLICATION OF THE DEMOGRAPHIC PRISONER'S DILEMMA
    Copyright (C) <2009>  <Wolfgang Radax and Bernhard Rengs>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Corresponding author: Wolfgang Radax <wolfgang.radax@econ.tuwien.ac.at>
 */

package demographicpd;

import java.util.List;
import java.util.ArrayList;

import uchicago.src.sim.engine.SimpleModel;
import uchicago.src.sim.engine.SimInit;
import uchicago.src.sim.space.Object2DTorus;

import uchicago.src.sim.util.Random;
import uchicago.src.sim.util.SimUtilities;

/**
 * DemographicPD is the central class of this replication and contains large parts
 * of the model logic (the rest of the model logic being located in {@link Agent}).
 * @author Wolfgang Radax, Vienna University of Technology
 */
public class DemographicPD extends SimpleModel {

    protected Object2DTorus     space;
    protected AgentList         agents;
    private PrisonersDilemma    pd;
    
    private int                 spaceXSize;
    private int                 spaceYSize;
    private int                 numAgents;
    
    private int                 initResources;
    private int                 birthResources;
    private int                 defectorRatio;

    private int                 maxAge;
    private int                 birthMinAge;
    private int                 birthMaxAge;
    
    private int                 pdReward;
    private int                 pdTemptation;
    private int                 pdPunishment;
    private int                 pdSuckers;
    
    private int                 mutationProbability;
    
    protected int               runUntil;
    
    private int                 idCounter;
    
    private int                 newBorn;
    private int                 died;
    
    private boolean             asynchronous;
    private boolean             randomBirthAge;    
    private boolean             initialEndowmentInherited;
    private boolean             dieImmediately;
    private boolean             removeDeadImmediately;
    private boolean             cernRandom;
    private boolean             repastShuffle;
    
    private java.util.Random    randomGen;
    
    /**
     * Destroys all objects for data storage ("flushes the simulation") and sets
     * all parameters to default values. Is called automatically at model setup.
     */
    public void setup() {
        super.setup();
        
        name = "Demographic PD";
        
        // Reset data structures;
        space = null;
        agents = null;
        pd = null;
        
        // Set default parameters
        spaceXSize              =  30;
        spaceYSize              =  30;
        numAgents               = 100;
        
        initResources           =   6;
        birthResources          =  10;
        defectorRatio           =  50;

        maxAge                  = 100;
        birthMinAge             =   0;
        birthMaxAge             = maxAge;

        pdReward                =   5;
        pdTemptation            =   6;
        pdPunishment            =  -5;
        pdSuckers               =  -6;
        
        runUntil = 1000;
        
        idCounter = 0;
        
        mutationProbability = 0;
        
        newBorn = 0;
        died = 0;
        
        asynchronous = true;
        randomBirthAge = false;
        initialEndowmentInherited = true;
        dieImmediately = true;
        removeDeadImmediately = false;
        cernRandom = true;
        repastShuffle = true;
        
        randomGen = null;
    }
    
    /**
     * Builds all the objects for data storage, initializes the random number
     * generator, and distributes the agents randomly on the torus. Is called
     * automatically at model initialization.
     */
    public void buildModel() {
        super.buildModel();
        
        if (cernRandom) {
            Random.setSeed(System.currentTimeMillis());
            Random.createUniform();
        }
        else {
            randomGen = new java.util.Random(System.currentTimeMillis());
        }
        
        space = new Object2DTorus(spaceXSize, spaceYSize);
        agents = new AgentList();
        
        pd = new PrisonersDilemma(pdReward, pdTemptation, pdPunishment, pdSuckers);
        
        int x, y;
        
        for (int i = 0; i < numAgents; i++) {
            do {
                x = nextRandomIntFromTo(0, spaceXSize - 1);
                y = nextRandomIntFromTo(0, spaceYSize - 1);
            } while (space.getObjectAt(x, y) != null);
            
            Agent agent = new Agent (idCounter, x, y, this);
            idCounter++;
            
            if (maxAge == 0) {
                agent.setAge(1);
            }
            else {
                agent.setAge(nextRandomIntFromTo(1, maxAge));
            }

            agents.add(agent);
            space.putObjectAt(x, y, agent);
        }
        
        for (Agent a : agents) {
            int mutation = nextRandomIntFromTo(1, 100);
            if (mutation <= defectorRatio) {
                a.setBehavior(PrisonersDilemma.DEFECT);
            }
        }
    }

    /**
     * Contains the sequence of events of the demographic prisoner's dilemma model:
     * move, play, give birth, die of age, age. Is called automatically every time
     * step.
     */
    public void step() {
        super.step();
        
        ArrayList<Agent> birthQueue = new ArrayList();
        newBorn = 0;
        died = 0;
        
        if (asynchronous) {
            for (Agent a : agents) {

                //move
                a.move();

                //play against all four neighbors (at maximum)
                for (Agent n : a.getNeighbors()) {
                    if (n.isAlive() && !n.isNewBorn()) {
                        a.play(n);
                    }
                }

                //give birth
                Agent newAgent = a.giveBirth(idCounter);
                if (newAgent != null) {
                    idCounter++;
                    space.putObjectAt(newAgent.getX(), newAgent.getY(), newAgent);
                    birthQueue.add(newAgent);
                }

                //die of age
                if (a.isAlive() && a.getAge() >= maxAge && maxAge > 0) {
                    a.setAlive(false);
                    if (removeDeadImmediately) {
                        space.putObjectAt(a.getX(), a.getY(), null);
                    }
                }
            }
        }
        else {
            //all agents move
            for (Agent a : agents) {
                a.move();
            }

            //all agents play
            for (Agent a : agents) {
                //play against all four neighbors (at maximum)
                for (Agent n : a.getNeighbors()) {
                    if (n.isAlive() && !n.isNewBorn()) {
                        a.play(n);
                    }
                }
            }

            //all agents give birth
            for (Agent a : agents) {
                Agent newAgent = a.giveBirth(idCounter);
                if (newAgent != null) {
                    idCounter++;
                    space.putObjectAt(newAgent.getX(), newAgent.getY(), newAgent);
                    birthQueue.add(newAgent);
                }
            }

            //all agents die of age
            for (Agent a: agents) {
                //die of age
                if (a.isAlive() && a.getAge() >= maxAge && maxAge > 0) {
                    a.setAlive(false);
                    if (removeDeadImmediately) {
                        space.putObjectAt(a.getX(), a.getY(), null);
                    }
                }
            }
        }
        newBorn = birthQueue.size();

        //add newborn agents to the agents-list
        for (Agent a : birthQueue) {
            agents.add(a);
            a.setNewBorn(false);
            
        }
            
        List<Agent> dieAgents = new ArrayList();
        dieAgents.addAll(agents);
        
        //remove died agents from the agents-list
        for (Agent a : dieAgents) {
            if (!a.isAlive()) {
                agents.remove(a);
                if (!removeDeadImmediately) {
                    space.putObjectAt(a.getX(), a.getY(), null);
                }
                died++;
            }
        }
        
        //let all agents age by 1
        for (Agent a: agents) {
            a.setAge(a.getAge() + 1);
        }

        shuffle(agents);

    }
    
    /**
     * Returns a random int number between <code>from</code> and <code>to</code>
     * (including both values) drawn from a uniform distribution. Depending on
     * the chosen random number generator, the number is generated by the CERN
     * random-package provided by Repast or by the java.util.Random library
     * provided by the Java framework.
     * @param from  Min value of the random number to be returned
     * @param to    Max value of the random number to be returned
     * @return      Random int number between <code>from</code> and <code>to</code>
     *              drawn from a uniform distribution
     * @see #setCernRandom
     */
    public int nextRandomIntFromTo(int from, int to) {
        int retValue = 0;
        
        if (cernRandom) {
            retValue = Random.uniform.nextIntFromTo(from, to);
        }
        else {
            retValue = from + randomGen.nextInt(to - from + 1);
        }
        
        return retValue;
    }       
    
    /**
     * Shuffles the elements of the passed ArrayList to prevent artefacts in
     * the results. Depending on the chosen shuffling method, either the
     * shuffle-method provided by Repast (<code>SimUtilities.shuffle</code>) or 
     * the algorithm given in Epstein (1999) is used.
     * @param l The ArrayList to be shuffled
     * @see #setRepastShuffle
     */
    public void shuffle(ArrayList l) {
        if (repastShuffle) {
            SimUtilities.shuffle(l);
        }
        else {
            int n = l.size() / 2;
            for (int i = 0; i < n; i++) {
                int o1 = nextRandomIntFromTo(0, l.size() - 1);
                int o2 = nextRandomIntFromTo(0, l.size() - 1);

                Object o = l.get(o1);
                l.set(o1, l.get(o2));
                l.set(o2, o);
            }
        }
    }
    
    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        // TODO code application logic here
        
        SimInit init = new SimInit();
        DemographicPD dpd = new DemographicPD();
        init.loadModel(dpd, null, false);
    }
    
    //SETTERS AND GETTERS
    public Object2DTorus getSpace() {
        return this.space;
    }
    
    public PrisonersDilemma getPrisonersDilemma() {
        return this.pd;
    }
    
    public int getSpaceXSize() {
        return spaceXSize;
    }
    
    /**
     * Determines the x-size of the torus.
     * @param spaceXSize X-size of the torus
     */
    public void setSpaceXSize(int spaceXSize) {
        this.spaceXSize = spaceXSize;
    }
    
    public int getSpaceYSize() {
        return spaceYSize;
    }
    
    /**
     * Determines the y-size of the torus.
     * @param spaceYSize Y-Size of the torus.
     */
    public void setSpaceYSize(int spaceYSize) {
        this.spaceYSize = spaceYSize;
    }
    
    public int getInitResources() {
        return this.initResources;
    }
    
    /**
     * Determines the amount of resources each agent possesses at the start of
     * the simulation or at birth ("initial endowment").
     * @param initResources Amount of resources each agent possesses at the start
     *                      the simulation
     */
    public void setInitResources(int initResources) {
        this.initResources = initResources;
    }
    
    public int getNumAgents() {
        return this.numAgents;
    }
    
    /**
     * Determines the number of agents at the start of the simulation.
     * @param numAgents Number of agents at the start of the simulation
     */
    public void setNumAgents(int numAgents) {
        this.numAgents = numAgents;
    }
    
    public int getMaxAge() {
        return this.maxAge;
    }
    
    /**
     * Determines the maximum age an agent can reach before he dies. If
     * <code>maxAge</code> is set to 0, no maximum age is imposed, i.e. agents
     * can only die as a result of negative payoffs from the games they played.
     * @param maxAge Maximum age an agent can reach
     */
    public void setMaxAge(int maxAge) {
        this.maxAge = maxAge;
    }

    public int getPdReward() {
        return this.pdReward;
    }
    
    /**
     * Determines the value for payoff "Reward" in the prisoner's dilemma.
     * @param pdReward Value for payoff "Reward" in the prisoner's dilemma
     */
    public void setPdReward(int pdReward) {
        this.pdReward = pdReward;
    }
    
    public int getPdPunishment() {
        return this.pdPunishment;
    }
    
    /**
     * Determines the value for payoff "Punishment" in the prisoner's dilemma.
     * @param pdPunishment Value for payoff "Punishment" in the prisoner's dilemma
     */
    public void setPdPunishment(int pdPunishment) {
        this.pdPunishment = pdPunishment;
    }
    
    public int getPdTemptation() {
        return this.pdTemptation;
    }
    
    /**
     * Determines the value for payoff "Temptation" in the prisoner's dilemma.
     * @param pdTemptation Value for payoff "Temptation" in the prisoner's dilemma
     */
    public void setPdTemptation(int pdTemptation) {
        this.pdTemptation = pdTemptation;
    }
    
    public int getPdSuckers() {
        return this.pdSuckers;
    }
    
    /**
     * Determines the value for payoff "Sucker's Payoff" in the prisoner's dilemma.
     * @param pdSuckers Value for payoff "Sucker's Payoff" in the prisoner's dilemma
     */
    public void setPdSuckers(int pdSuckers) {
        this.pdSuckers = pdSuckers;
    }
    
    public int getMutationProbability() {
        return this.mutationProbability;
    }
    
    /**
     * Determines the probability of mutation in percent. That is, a value between
     * 0 and 100 has to be provided. 
     * @param mutationProbability Probability of mutation (in percent)
     */
    public void setMutationProbability(int mutationProbability) {
        this.mutationProbability = mutationProbability;
    }
    
    public int getBirthMinAge() {
        return this.birthMinAge;
    }
    
    /**
     * Determines the minimum age an agent has to reach in order to be able to 
     * give birth to offspring. If <code>birthMinAge</code> is set to 0, no
     * minimum birth age is imposed. <p> Please note that this is an extension to 
     * the original Epstein (1999) model and was not part of that original model.
     * For matters of replication the value was always set to 0 (no minimum birth
     * age).
     * @param birthMinAge Minimum age in order to be able to give birth to offspring
     */
    public void setBirthMinAge(int birthMinAge) {
        this.birthMinAge = birthMinAge;
    }
    
    public int getBirthMaxAge() {
        return this.birthMaxAge;
    }
    
    /**
     * Determines the maximum age an agent may reach in order to be able to give
     * birth to offspring. If <code>birthMaxAge</code> is set to 0, no maximum
     * birth age is imposed. <p> Please note that this is an extension to 
     * the original Epstein (1999) model and was not part of that original model.
     * For matters of replication the value was always set to 0 (no maximum birth
     * age).
     * @param birthMaxAge   Maximum age in order to still be able to give birth to
     *                      offspring
     */
    public void setBirthMaxAge(int birthMaxAge) {
        this.birthMaxAge = birthMaxAge;
    }
    
    public boolean getRandomBirthAge() {
        return this.randomBirthAge;
    }
    
    public int getBirthResources() {
        return this.birthResources;
    }
    
    /**
     * Determines the minimum amount of resources an agent has to possess in order
     * to give birth to offspring. 
     * @param birthResources Amount of resources required to give birth
     */
    public void setBirthResources(int birthResources) {
        this.birthResources = birthResources;
    }
    
    /**
     * Determines whether agents are born with age 0 or with a random age between
     * 0 and <code>maxAge - 1</code>.
     * @param randomBirthAge Determination of birth age
     */
    public void setRandomBirthAge(boolean randomBirthAge) {
        this.randomBirthAge = randomBirthAge;
    }
    
    public int getDefectorRatio() {
        return this.defectorRatio;
    }
    
    /**
     * Determines the ratio of defectors in the initial agent population in
     * percent. That is, a value between 0 and 100 has to be provided. 
     * @param defectorRatio Ratio of defectors in the initial agent population
     *                      (in percent)
     */
    public void setDefectorRatio(int defectorRatio) {
        this.defectorRatio = defectorRatio;
    }
    
    public boolean getAsynchronous() {
        return this.asynchronous;
    }
    
    /**
     * Determines the mode of updating. If <code>asynchronous</code> is set to
     * <code>true</code>, asynchronous updating is used, otherwise synchronous
     * updating is employed.
     * @param asynchronous Mode of updating
     */
    public void setAsynchronous(boolean asynchronous) {
        this.asynchronous = asynchronous;
    }
    
    public boolean getInitialEndowmentInherited() {
        return this.initialEndowmentInherited;
    }
    
    /**
     * Determines whether the initial endowment an agent receives at birth is 
     * inherited from his parent (<code>initialEndowmentInherited</code> set to
     * <code>true</code>) or the initial endowment is provided by nature.
     * @param initialEndowmentInherited Determines the origin of the initial endowment.
     */
    public void setInitialEndowmentInherited(boolean initialEndowmentInherited) {
        this.initialEndowmentInherited = initialEndowmentInherited;
    }
    
    public boolean getDieImmediately() {
        return this.dieImmediately;
    }
    
    /**
     * Determines whether agents die immediately after their resources fall below 0,
     * or only when it's their turn. This concerns those agents who play games when
     * it's not their turn, i.e. those agents who are the playing partners of the
     * currently active agent.
     * @param dieImmediately Moment of death when resources fall below 0
     */
    public void setDieImmediately(boolean dieImmediately) {
        this.dieImmediately = dieImmediately;
    }
    
    public boolean getRemoveDeadImmediately() {
        return this.removeDeadImmediately;
    }
    
    /**
     * Determines whether dead agents are removed from the torus at the very instant
     * their resources fall below 0 or later collectively at the end of each time step.
     * @param removeDeadImmediately Moment of removal from the torus
     */
    public void setRemoveDeadImmediately(boolean removeDeadImmediately) {
        this.removeDeadImmediately = removeDeadImmediately;
    }
    
    public boolean getCernRandom() {
        return this.cernRandom;
    }
    
    /**
     * Determines the random number generator. <code>cernRandom</code> set to
     * <code>true</code> corresponds to the CERN random number generator provided
     * by Repast. Otherwise java.util.Random is used.
     * @param cernRandom Random number generator to be used throughout the simulation
     * @see #nextRandomIntFromTo
     */
    public void setCernRandom(boolean cernRandom) {
        this.cernRandom = cernRandom;
    }
    
    public boolean getRepastShuffle() {
        return this.repastShuffle;
    }
    
    /**
     * Determines the method of list shuffling. <code>repastShuffle</code> set to
     * <code>true</code> corresponds to the shuffling-method provided by Repast
     * (<code>SimUtilities.shuffle</code>). Otherwise the shuffling-method
     * suggested by Epstein (1999) is employed.
     * @param repastShuffle Method of list shuffling
     * @see #shuffle 
     */
    public void setRepastShuffle(boolean repastShuffle) {
        this.repastShuffle = repastShuffle;
    }
    
    public double getNewBorn() {
        return this.newBorn;
    }
    
    public double getDied() {
        return this.died;
    }
            
}
