/*
    REPLICATION OF THE DEMOGRAPHIC PRISONER'S DILEMMA
    Copyright (C) <2009>  <Wolfgang Radax and Bernhard Rengs>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
    Corresponding author: Wolfgang Radax <wolfgang.radax@econ.tuwien.ac.at>
 */

package demographicpd;

import uchicago.src.sim.engine.SimInit;
import uchicago.src.sim.engine.Controller;
import uchicago.src.sim.analysis.LocalDataRecorder;

/**
 * DemographicPDBatch implements a No-GUI-version of the Demographic Prisoner's
 * Dilemma for multiple ("batch") runs. Input parameters are provided by a
 * parameter-file, the output is written to an output-file ("data.txt").
 * @author Wolfgang Radax, Vienna University of Technology
 */
public class DemographicPDBatch extends DemographicPD {

    private LocalDataRecorder recorder;
    
    /**
     * Default constructor.
     */
    public DemographicPDBatch() {
        Controller.ALPHA_ORDER = false;
    }
    
    /**
     * Calls garbage collector and specifies parameters which can be set by
     * input-file. "Flushes the simulation". Is called automatically at model
     * setup.
     */
    public void setup() {
        super.setup();
        
        this.isGui = false;
        
        System.gc();
        params = new String[]{
                "seed",
                "runUntil",
                "spaceXSize",
                "spaceYSize",
                "initResources",
                "numAgents",
                "pdReward",
                "pdTemptation",
                "pdSuckers",
                "pdPunishment",
                "defectorRatio",
                "maxAge",
                "birthMinAge",
                "birthMaxAge",
                "birthResources",
                "mutationProbability",
                "asynchronous",
                "randomBirthAge",                
                "initialEndowmentInherited",
                "dieImmediately",
                "removeDeadImmediately",
                "cernRandom",
                "repastShuffle"
            };
    }
    
    /**
     * Builds the objects required for collecting data and writing the output-file.
     * Is called automatically at model initialization.
     */
    public final void buildModel() {
        super.buildModel();
        
        setStoppingTime(runUntil);
        
        recorder = new LocalDataRecorder("./data.txt", this);
        recorder.setDelimeter(";");
        
        recorder.createNumericDataSource("currentdefratio", this, "computeCurrentDefRatio");
        recorder.createNumericDataSource("numcooperators", this, "computeNumCooperators");
        recorder.createNumericDataSource("numdefectors", this, "computeNumDefectors");
        recorder.createNumericDataSource("avgallage", this, "computeAvgAllAge");
        recorder.createNumericDataSource("avgdefage", this, "computeAvgDefAge");
        recorder.createNumericDataSource("avgcoopage", this, "computeAvgCoopAge");
        recorder.createNumericDataSource("avgallres", this, "computeAvgAllRes");
        recorder.createNumericDataSource("avgdefres", this, "computeAvgDefRes");
        recorder.createNumericDataSource("avgcoopres", this, "computeAvgCoopRes");
    }
    
    /**
     * Collects the data for output in the final period (specified by parameter
     * "runUntil" in the output-file). Is called automatically each time step.
     */
    public final void step() {
        super.step();
        if (getTickCount() == runUntil) {
            recorder.record();
        }
    }
    
    /**
     * Writes collected data to the output-file. Is called automatically after
     * the last time step.
     */
    public final void atEnd() {
        super.atEnd();
        recorder.write();
    }
    
    /**
     * Encapsulates method {@link AgentList#getAvgResources}.
     * @return Average resources of population
     */
    public final int computeAvgAllRes() {
        return (int) agents.getAvgResources();
    }
    
    /**
     * Encapsulates method {@link AgentList#getAvgDefectorResources}.
     * @return Average resources of all defectors
     */
    public final int computeAvgDefRes() {
        return (int) agents.getAvgDefectorResources();
    }
    
    /**
     * Encapsulates method {@link AgentList#getAvgCooperatorResources}.
     * @return Average resources of all cooperators
     */
    public final int computeAvgCoopRes() {
        return (int) agents.getAvgCooperatorResources();
    }
    
    /**
     * Encapsulates method {@link AgentList#getCurrentDefectorRatio}.
     * @return Share of defectors in the population
     */
    public final int computeCurrentDefRatio() {
        return (int) agents.getCurrentDefectorRatio();
    }
    
    /**
     * Encapsulates method {@link AgentList#getAvgAge}.
     * @return Average age of the population
     */
    public final int computeAvgAllAge() {
        return (int) agents.getAvgAge();
    }
    
    /**
     * Encapsulates method {@link AgentList#getAvgDefectorAge}.
     * @return Average age of all defectors
     */
    public final int computeAvgDefAge() {
        return (int) agents.getAvgDefectorAge();
    }
    
    /**
     * Encapsulates method {@link AgentList#getAvgCooperatorAge}.
     * @return Average age of all cooperators
     */
    public final int computeAvgCoopAge() {
        return (int) agents.getAvgCooperatorAge();
    }
    
    /**
     * Encapsulates method {@link AgentList#getNumberCooperators}.
     * @return Number of cooperators
     */
    public final int computeNumCooperators() {
        return agents.getNumberCooperators();
    }
    
    /**
     * Encapsulates method {@link AgentList#getNumberDefectors}.
     * @return Number of defectors
     */
    public final int computeNumDefectors() {
        return agents.getNumberDefectors();
    }
    
    public static void main(final String[] args) {
        final SimInit init = new SimInit();
        final DemographicPDBatch m = new DemographicPDBatch();
        init.loadModel(m, "./params.txt", true);
    }

    //SETTERS AND GETTERS
    public final int getRunUntil() {
        return runUntil;
    }
    
    public final void setRunUntil(final int v) {
        runUntil = v;
    }
}
