/*
    REPLICATION OF THE DEMOGRAPHIC PRISONER'S DILEMMA
    Copyright (C) <2009>  <Wolfgang Radax and Bernhard Rengs>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Corresponding author: Wolfgang Radax <wolfgang.radax@econ.tuwien.ac.at>
 */

package demographicpd;

import uchicago.src.sim.engine.SimInit;
import uchicago.src.sim.engine.Controller;
import uchicago.src.sim.gui.DisplaySurface;
import uchicago.src.sim.gui.DisplayConstants;
import uchicago.src.sim.gui.Object2DDisplay;
import uchicago.src.sim.analysis.OpenSequenceGraph;
import uchicago.src.sim.analysis.OpenHistogram;
import uchicago.src.sim.analysis.BinDataSource;
import uchicago.src.sim.analysis.Plot;
import uchicago.src.sim.analysis.Sequence;

import java.awt.Color;
import java.util.List;

/**
 * DemographicPDGUI implements the graphical ("GUI") version of the Demographic Prisoner's Dilemma replication.
 * Inherits model logic from {@link DemographicPD} and serves as a view for the latter following a simplified
 * MVC pattern.
 * @author Wolfgang Radax, Vienna University of Technology
 */
public class DemographicPDGUI extends DemographicPD {
    
    private DisplaySurface dispsurf;
    
    private OpenSequenceGraph   defectorRatioGraph;
    private OpenSequenceGraph   newBornAndDiedGraph;
    private OpenSequenceGraph   avgAgeGraph;
    private Plot                phaseDiagram;
    
    private OpenHistogram bar;
    
    BinDataSource source = new BinDataSource()  {
      public double getBinValue(Object o) {
        Agent agent = (Agent)o;
        return agent.getAge();
      }
    };
    
    final class NumDefectors implements Sequence {
        public final double getSValue() {
            return (double) agents.getNumberDefectors();
        }
    }
    
    final class NumCooperators implements Sequence {
        public final double getSValue() {
            return (double) agents.getNumberCooperators();
        }
    }

    final class AvgAgeSeq implements Sequence {
        public final double getSValue() {
            return (double) agents.getAvgAge();
        }
    }
    
    final class AvgDefectorAgeSeq implements Sequence {
        public final double getSValue() {
            return (double) agents.getAvgDefectorAge();
        }
    }
    
    final class AvgCooperatorAgeSeq implements Sequence {
        public final double getSValue() {
            return (double) agents.getAvgCooperatorAge();
        }
    }
    
    final class NumNewBornSeq implements Sequence {
        public final double getSValue() {
            return getNewBorn();
        }
    }
    
    final class NumDiedSeq implements Sequence {
        public final double getSValue() {
            return getDied();
        }
    }
    
    /**
     * Default constructor.
     */
    public DemographicPDGUI () {
        Controller.ALPHA_ORDER = false;
    }
    
    /**
     * Destroys all objects for graphical representation and calls garbage collector,
     * "flushes the simulation". Is called automatically at model setup (button "setup").
     */
    public final void setup() {
        super.setup();
        
        params = new String[]{
                "spaceXSize",
                "spaceYSize",
                "initResources",
                "numAgents",
                "pdReward",
                "pdTemptation",
                "pdSuckers",
                "pdPunishment",
                "defectorRatio",
                "maxAge",
                "birthMinAge",
                "birthMaxAge",
                "birthResources",
                "mutationProbability",
                "asynchronous",
                "randomBirthAge",                
                "initialEndowmentInherited",
                "dieImmediately",
                "removeDeadImmediately",
                "cernRandom",
                "repastShuffle"
            };
        
        if (dispsurf != null) {
            dispsurf.dispose();
        }
        dispsurf = null;
        
        if (this.bar != null) {
            bar.dispose();
        }
        bar = null;
        
        if (defectorRatioGraph != null) {
            defectorRatioGraph.dispose();
        }
        defectorRatioGraph = null;
        
        if (newBornAndDiedGraph != null) {
            newBornAndDiedGraph.dispose();
        }
        newBornAndDiedGraph = null;
        
        if (avgAgeGraph != null) {
            avgAgeGraph.dispose();
        }
        avgAgeGraph = null;
        
        if (phaseDiagram != null) {
            phaseDiagram.dispose();
        }
        phaseDiagram = null;
        
        System.gc();

        dispsurf = new DisplaySurface(this, "Demographic PD Display");
        registerDisplaySurface("Main", dispsurf);

        DisplayConstants.CELL_WIDTH = 30;
        DisplayConstants.CELL_HEIGHT = 30;
        
        System.out.println("Demographic PD v1.0");
        System.out.println("");
    }
    
    /**
     * Builds all the objects for graphical representation: Display for the torus,
     * graphs, phase diagram. Is called automatically at model initialization (button "Initialize").
     */
    public final void buildModel() {
        super.buildModel();
        
        final Object2DDisplay display = new Object2DDisplay(space);
        display.setObjectList((List) agents);
        dispsurf.addDisplayableProbeable(display, "Display");
        addSimEventListener(dispsurf);
        dispsurf.display();
        
        defectorRatioGraph = new OpenSequenceGraph("Population Size", this);
        defectorRatioGraph.setXRange(0, 50);
        defectorRatioGraph.setYRange(0.0, 1.0);
        defectorRatioGraph.setAxisTitles("Time", "Population Size");
        defectorRatioGraph.addSequence("Defectors", new NumDefectors(), Color.RED);
        defectorRatioGraph.addSequence("Cooperators", new NumCooperators(), Color.GREEN);
        defectorRatioGraph.display();
        defectorRatioGraph.step();
        
        newBornAndDiedGraph = new OpenSequenceGraph("Born and died", this);
        newBornAndDiedGraph.setXRange(0, 50);
        newBornAndDiedGraph.setYRange(0.0, 1.0);
        newBornAndDiedGraph.setAxisTitles("Time", "Born and died");
        newBornAndDiedGraph.addSequence("New Born", new NumNewBornSeq(), Color.GREEN);
        newBornAndDiedGraph.addSequence("Died", new NumDiedSeq(), Color.RED);
        newBornAndDiedGraph.display();
        newBornAndDiedGraph.step();

        avgAgeGraph = new OpenSequenceGraph("Avg. Age", this);
        avgAgeGraph.setXRange(0, 50);
        avgAgeGraph.setYRange(0.0, 1.0);
        avgAgeGraph.setAxisTitles("Time", "Avg. Age");
        avgAgeGraph.addSequence("Age", new AvgAgeSeq(), Color.BLACK);
        avgAgeGraph.addSequence("DefAge", new AvgDefectorAgeSeq(), Color.RED);
        avgAgeGraph.addSequence("CoopAge", new AvgCooperatorAgeSeq(), Color.GREEN);
        avgAgeGraph.display();
        avgAgeGraph.step();
        
        phaseDiagram = new Plot("Phase Diagram");
        phaseDiagram.setAxisTitles("Number Cooperators", "Number Defectors");
        phaseDiagram.display();
        phaseDiagram.setConnected(true);
        phaseDiagram.plotPoint(agents.getNumberCooperators(),
                agents.getNumberDefectors(), 1);
        phaseDiagram.updateGraph();
        phaseDiagram.setXRange(0, getSpaceXSize() * getSpaceYSize());
        phaseDiagram.setYRange(0, getSpaceXSize() * getSpaceYSize());
    }
    
    /**
     * Updates all graphical objects. Is called automatically every time step (buttons "Step" or "Start").
     */
    public final void step() {
        super.step();
        
        if (getTickCount() > 0 && bar == null && getMaxAge() > 0 && agents.size() > 0) {
            bar = new OpenHistogram("Agent Age Distribution", 10, 0);
            bar.setXRange(0, 100.0);
            bar.setYRange(0, 100.0);
            bar.createHistogramItem("Age", agents, source);
            bar.display();
            bar.step();
        }
        else if (bar != null && getMaxAge() > 0 && agents.size() > 0) {
            bar.step();
        }
        
        dispsurf.updateDisplay();

        defectorRatioGraph.step();        
        newBornAndDiedGraph.step();
        avgAgeGraph.step();
        phaseDiagram.plotPoint(agents.getNumberCooperators(),
                agents.getNumberDefectors(), 1);
        phaseDiagram.updateGraph();
    }
    
    public static void main(final String[] args) {
        SimInit init = new SimInit();
        
        DemographicPDGUI dpdgui = new DemographicPDGUI();
        init.loadModel(dpdgui, null, false);
    }
}
