/*
    REPLICATION OF THE DEMOGRAPHIC PRISONER'S DILEMMA
    Copyright (C) <2009>  <Wolfgang Radax and Bernhard Rengs>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
    Corresponding author: Wolfgang Radax <wolfgang.radax@econ.tuwien.ac.at>
 */

package demographicpd;

/**
 * PrisonersDilemma serves as a storage for PD-payoffs and returns the payoff
 * given the chosen strategies by the players.
 * @author Wolfgang Radax, Vienna University of Technology
 */
public class PrisonersDilemma {

    private int reward;
    private int temptation;
    private int punishment;
    private int suckers;
    
    public static final int COOPERATE = 0;
    public static final int DEFECT = 1;
    
    /**
     * Default constructor. There is no possibility to create a PD object without
     * specifying the payoff parameters.
     * @param R "Reward". Payoff in the case of both players cooperating.
     * @param T "Temptation". Payoff in the case of defection when the opponent cooperates.
     * @param P "Punishment". Payoff in the case of both players defecting.
     * @param S "Sucker's Payoff". Payoff in the case of cooperation when the opponent defects.
     */
    public PrisonersDilemma(int R, int T, int P, int S) {
        reward = R;
        temptation = T;
        punishment = P;
        suckers = S;
    }
    
    /**
     * Returns the respective payoff-value (R, T, P or S) for the "player" given
     * the strategies of the "player" and the "opponent" passed to this method.
     * @param playerStrategy    COOPERATE or DEFECT
     * @param opponentStrategy  COOPERATE or DEFECT
     * @return Payoff for "player" given the strategies of the two players "player" and "opponent"
     */
    public int getPayoff(int playerStrategy, int opponentStrategy) {
        if (playerStrategy == COOPERATE && opponentStrategy == COOPERATE) {
            return reward;
        }
        
        if (playerStrategy == COOPERATE && opponentStrategy == DEFECT) {
            return suckers;
        }
        
        if (playerStrategy == DEFECT && opponentStrategy == COOPERATE) {
            return temptation;
        }
        
        if (playerStrategy == DEFECT && opponentStrategy == DEFECT) {
            return punishment;
        }
        
        return 0;
    }
}
